#include "hash.h"
#include <stdlib.h>
#include <string.h>

/*
 * Private Interface.
 */

#define HASHSIZE    101

struct Nlist {
    char *word;         /* search word */
    char *defn;         /* word definition */
    struct Nlist *next; /* pointer to next entry in chain */
};

struct Dictionary {
    struct Nlist *table[HASHSIZE]; /* table is an array of pointers to entries */
};

static unsigned hash_function(const char *str)
/* Hashing function converts a string to an index within hash table. */
{
    const int HashValue = 31;
    unsigned h;

    for (h = 0; *str != '\0'; ++str)
        h = *str + HashValue * h;
    return h % HASHSIZE;
}

static char *allocate_string(const char *str)
/* Dynamically allocate a duplicate of str. Return NULL if allocation fails. */
{
    char *tmp = (char *)malloc(strlen(str)+1);
    if (tmp) 
        strcpy(tmp, str);
    return tmp;
}

static void destroynode(struct Nlist *n)
/* Release memory allocated for node (including internals) */
{
    free(n->word);
    free(n->defn);
    free(n);
}

static struct Nlist* makenode(const char *key, const char *defn)
/* Allocate memory for a new node and its contents */
{
    struct Nlist *n = (struct Nlist *)malloc(sizeof(struct Nlist));
    if (n == NULL) 
        return NULL;

    n->next = NULL;
    n->word = allocate_string(key);
    n->defn = allocate_string(defn);
    if (n->word == NULL || n->defn == NULL) {
        destroynode(n);
        return NULL;
    }
    return n;
}

/*
 * Implementations for Public Interface.
 */

Dictionary *create_table(void)
/* Create dictionary and set all its table entries to NULL */
{
    unsigned i;
    Dictionary *d = (Dictionary *)malloc(sizeof(Dictionary));

    if (d)
        for (i = 0; i < HASHSIZE; ++i)
            d->table[i] = NULL;
    return d;
}

void destroy_table(Dictionary *d)
/* Iterate through entire table freeing contents, then delete dictionary itself. */
{
    unsigned i;

    for (i = 0; i < HASHSIZE; ++i) {
        struct Nlist *pnode = d->table[i];

        while (pnode) {
            struct Nlist *tmp = pnode;
            pnode = pnode->next;
            destroynode(tmp);
        }
    }
    free(d);
}

int add_word(Dictionary *dict, const char *key, const char *defn)
/* Add new word to table. Replaces old definition if word already exists.
 * Return 0 if successful, and -1 is fails. */
{
    unsigned i = hash_function(key); /* get table index */
    struct Nlist *pnode = dict->table[i];             

    while (pnode && strcmp(pnode->word, key) != 0) /* search chain */
        pnode = pnode->next;

    if (pnode) { /* match found, replace definition */
        char *str = allocate_string(defn);
        if (str == NULL) /* allocation fails, return fail and keep old defn */
            return -1;

        free(pnode->defn);
        pnode->defn = str;
    }
    else { /* no match, add new entry to head of chain */
        pnode = makenode(key, defn);
        if (pnode == NULL)
            return -1;

        pnode->next = dict->table[i];
        dict->table[i] = pnode;
    }
    return 0;
}

char *find_word(const Dictionary *dict, const char *key)
/* Find definition for keyword. Return NULL if key not found. */
{
    unsigned i = hash_function(key); /* get table index */
    struct Nlist *pnode = dict->table[i];             

    while (pnode && strcmp(pnode->word, key) != 0) /* search index chain */
        pnode = pnode->next;

    if (pnode) /* match found */
        return pnode->defn;
    return NULL;
}

void delete_word(Dictionary *dict, const char *key)
/* Remove node associated with keyword. If key is not found, do nothing. */
{
    unsigned i = hash_function(key); /* get table index */
    struct Nlist *pnode = dict->table[i];
    if (pnode == NULL)
        return;

    if (strcmp(pnode->word, key) == 0) { /* special case: match head node */
        dict->table[i] = pnode->next;
        destroynode(pnode);
        return;
    }

    while (pnode->next && strcmp(pnode->next->word, key) != 0) /* search index chain */
        pnode = pnode->next;

    if (pnode->next) { /* match found */
        struct Nlist *n = pnode->next;
        pnode->next = pnode->next->next; /* cut out link */
        destroynode(n);
    }
}
